/*
 * WMMail - Window Maker Mail
 *
 * Copyright (c) 1996, 1997, 1998  Per Liden
 * Copyright (c) 1997, 1998  Bryan Chan
 * Copyright (c) 2003 Alberto Gonz'alez Palomo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * sylpheed.c: functions to handle Sylpheed-style mail folders
 *
 * Version 1.0, 2003-02-01
 * Modified from mh.c version 1.1, following the code in GKrellM
 * Mailwatch plugin version 2.1, function "check_sylpheed_maildir()".
 *
 * GKrellM Mailwatch: http://gkrellm.luon.net/mailwatch.phtml
 * Sylpheed: http://sylpheed.good-day.net/
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <time.h>
#include <unistd.h>
#include "wmmail.h"
#include "wmutil.h"

#ifdef SYLPHEED_SUPPORT

/* 
 * Supported key(s) in the Options dictionary:
 *
 *   Path
 *
 * This module contains routines to check a Sylpheed-style (similar
 * to MH) mail folder during the periodic update of wmmail_status.
 * We can classify a given mail folder as being in one of the states
 * listed below.
 *
 * The states and their corresponding effects on wmmail_status are:
 *
 *   - no folder or empty folder                NO_MAIL
 *   - no new mails                             OLD_MAIL
 *   - same number of new mails or fewer        no change
 *   - more new mails than last time            NEW_MAIL
 *
 * We assume that overriding automatically changes wmmail_status
 * to OLD_MAIL.
 *
 */

/* internal function */
static int count_mail(char *, int *, int *);


#define GET_ENTRY(x,y,z) { proplist_t sub_key = PLMakeString(z); \
                           y = PLGetDictionaryEntry(x, sub_key); \
                           PLRelease(sub_key); }

int SYLPHEED_check(Mailbox *mailbox, int *beep, int *redraw, int *run)
{
  proplist_t  path;
  char       *mailbox_path;
  int         prev_status, 
              prev_new_mail_count;

  prev_status = mailbox->status;
  prev_new_mail_count = mailbox->new_mail_count;

  GET_ENTRY(mailbox->options, path, "Path");
  if (path == NULL)
  {
    croak("mailbox \"%s\" missing option \"Path\"; ignored", mailbox->name);
    return False;
  }
  else if (!PLIsString(path))
  {
    croak("mailbox \"%s\" has invalid path; ignored", mailbox->name);
    return False;
  }
  else
    mailbox_path = expand_path(PLGetString(path));

  if ( count_mail(mailbox_path,
                  &mailbox->total_mail_count,
                  &mailbox->new_mail_count)   )
  {
    if (mailbox->total_mail_count == 0)
    {
      /* there is no mail in the mailbox */
      mailbox->status = NO_MAIL;
    }
    else if (mailbox->new_mail_count == 0)
    {
      /* there are no new mails */
      mailbox->status = OLD_MAIL;
    }
    else if (mailbox->new_mail_count > prev_new_mail_count)
    {
      /* new mails have arrived! */
      if (mailbox->status == NEW_MAIL && always_new_mail_exec)
        *run |= True;
      else if (mailbox->status == NEW_MAIL)
        *run |= False;
      else
        *run |= True;

      *beep = True;

      mailbox->status = NEW_MAIL;
    }

    /* else no change */
  }
  else
  {
    /* no such mailbox */
    croak("error reading Sylpheed folder \"%s\"; ignored", mailbox->name);
    wfree(mailbox_path);
    return False;
  }

  *redraw |= (prev_status != mailbox->status);

  mailbox->last_update = time(NULL);

  wfree(mailbox_path);
  return True;
}


#define SYLPHEED_MESSAGE_IS_NEW    1
#define SYLPHEED_MESSAGE_IS_UNREAD 2
#define COUNT_UNREAD_AS_NEW False

static int count_mail(char *path, int *total_mail_count, int *new_mail_count)
{
  struct dirent *entry;
  DIR           *folder;
  FILE          *seq;
  char           buf[512];
  int            new_mail = 0,
                 all_mail = 0,
                 first_unseen = 0,
                 last_unseen = 0;
  int version;/* Read from ".sylpheed_mark": format version.        */
  int num;    /* Read from ".sylpheed_mark": message number.        */
  int flags;  /* Read from ".sylpheed_mark": message flags.         */
  int files_seen   = 0;/* Files mentioned in the ".sylpheed_mark" file.   */
  int files_in_dir = 0;/* Files physically in the directory (folder).     */

  /* Count total messages. */
  if ( !(folder = opendir(path)) )
  {
    croak("cannot open directory %s", path);
    return False;
  }

  while ((entry = readdir(folder)) != NULL)
  {
    char *tmp = entry->d_name;
    int   is_message = True;

    while (*tmp) { (void) (isdigit(*tmp++) || (is_message = False)); }

    if (is_message) { files_in_dir++; }
  }

  closedir(folder);

  sprintf(buf, "%s/.sylpheed_mark", path);

  /* find the new messages in the folder */
  if (seq = fopen(buf, "r"))
  {
    if (fread(&version, sizeof(version), 1, seq) != 1 || version != 2)
    {
	fclose(seq);
        croak(".sylpheed_mark format version not supported: %d", version);
        return False;
    }

    while (fread(&num, sizeof(num), 1, seq) == 1)
    {
	if (fread(&flags, sizeof(flags), 1, seq) != 1)
	    break;

	if (flags & SYLPHEED_MESSAGE_IS_NEW)
	    new_mail++;
	else if (COUNT_UNREAD_AS_NEW && flags & SYLPHEED_MESSAGE_IS_UNREAD)
	    new_mail++;

	files_seen++;
    }
    
    if (files_seen < files_in_dir)
      {
        /* This happens when new messages arrive and Sylpheed still hasn't
         * processed that folder, such as when it's not running.
         */
	new_mail += (files_in_dir - files_seen);
        all_mail = files_in_dir;
      }
    else
      {
        /* files_seen should never be greater than files_in_dir, unless
         * the .sylpheed_mark file gets corrupted. We take all_mail from
         * it, so that hopefully the user would notice if this number is
         * bigger than the number of actual message files.
         */
        all_mail = files_seen;
      }

    fclose(seq);
  }
  else
  {
    croak(".sylpheed_mark not found or unreadable in %s", path);
    return False;
  }

  *total_mail_count = all_mail;
  *new_mail_count = new_mail;

  return True;
}

#endif
